@@ -1,8 +1,10 @@
 /* SASL server API implementation
+ * Rob Siemborski
  * Tim Martin
+ * $Id: client.c,v 1.40 2002/01/10 22:13:45 rjs3 Exp $
  */
 /* 
- * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -47,212 +49,31 @@
 #include <limits.h>
 #include <ctype.h>
 #include <string.h>
-#include <sasl.h>
-#include <saslutil.h>
+
+/* SASL Headers */
+#include "sasl.h"
+#include "saslplug.h"
+#include "saslutil.h"
 #include "saslint.h"
 
-static int
-external_client_new(void *glob_context __attribute__((unused)),
-		    sasl_client_params_t *params __attribute__((unused)),
-		    void **conn_context)
-{
-  if (!params
-      || !params->utils
-      || !params->utils->conn
-      || !conn_context)
-    return SASL_BADPARAM;
-  if (!params->utils->conn->external.auth_id)
-    return SASL_NOMECH;
-  *conn_context = NULL;
-  return SASL_OK;
-}
-
-static int
-external_client_step(void *conn_context __attribute__((unused)),
-		     sasl_client_params_t *params,
-		     const char *serverin __attribute__((unused)),
-		     int serverinlen,
-		     sasl_interact_t **prompt_need,
-		     char **clientout,
-		     int *clientoutlen,
-		     sasl_out_params_t *oparams)
-{
-  int result;
-  sasl_getsimple_t *getuser_cb;
-  void *getuser_context;
-  const char *user;
-  unsigned len;
-
-  if (!params
-      || !params->utils
-      || !params->utils->conn
-      || !params->utils->getcallback
-      || !clientout
-      || !clientoutlen
-      || !oparams)
-    return SASL_BADPARAM;
-
-  if (!params->utils->conn->external.auth_id)
-    return SASL_BADPROT;
-
-  if (serverinlen != 0)
-    return SASL_BADPROT;
-
-  if (prompt_need && *prompt_need) {
-    /* Second time through; we used a SASL_INTERACT to get the user. */
-    if (! (*prompt_need)[0].result)
-      return SASL_BADPARAM;
-    user = (*prompt_need)[0].result;
-    *clientoutlen = strlen(user);
-    params->utils->free(*prompt_need);
-    *prompt_need = NULL;
-  } else {
-    /* We need to get the user. */
-    result = params->utils->getcallback(params->utils->conn,
-					SASL_CB_USER,
-					&getuser_cb,
-					&getuser_context);
-    switch (result) {
-    case SASL_INTERACT:
-      /* Set up the interaction... */
-      if (prompt_need) {
-	*prompt_need = params->utils->malloc(sizeof(sasl_interact_t) * 2);
-	if (! *prompt_need)
-	  return SASL_FAIL;
-	memset(*prompt_need, 0, sizeof(sasl_interact_t) * 2);
-	(*prompt_need)[0].id = SASL_CB_USER;
-	(*prompt_need)[0].prompt = "Authorization Identity";
-	(*prompt_need)[0].defresult = "";
-	(*prompt_need)[1].id = SASL_CB_LIST_END;
-      }
-      return SASL_INTERACT;
-    case SASL_OK:
-      if (getuser_cb &&
-	  (getuser_cb(getuser_context,
-		      SASL_CB_USER,
-		      &user,
-		      &len)
-	   == SASL_OK)) {
-	*clientoutlen = strlen(user);
-	break;
-      }
-      /* Otherwise, drop through into the default we-lose case. */
-    default:
-      /* Assume userid == authid. */
-      user = NULL;
-      *clientoutlen = 0;
-    }
-  }
-
-  *clientout = params->utils->malloc(*clientoutlen + 1);
-  if (! clientout)
-    return SASL_FAIL;
-  if (user)
-    memcpy(*clientout, user, *clientoutlen);
-  (*clientout)[*clientoutlen] = '\0';
-  
-  if (prompt_need)
-    *prompt_need = NULL;
-
-  result = _sasl_strdup(params->utils->conn->external.auth_id,
-			&oparams->authid,
-			NULL);
-  if (result == SASL_OK)
-  {
-    oparams->doneflag = 1;
-    oparams->mech_ssf = 0;
-    oparams->maxoutbuf = 0;
-    oparams->encode_context = NULL;
-    oparams->encode = NULL;
-    oparams->getmic = NULL;
-    oparams->decode_context = NULL;
-    oparams->decode = NULL;
-    oparams->verifymic = NULL;
-    oparams->user = NULL;
-    oparams->realm = NULL;
-    oparams->param_version = 0;
-
-    return SASL_OK;
-
-  }  /* otherwise */
-
-  sasl_FREE(*clientout);
-  *clientout = NULL;
-
-  return result;
-}
-
-static const sasl_client_plug_t external_client_mech = {
-  "EXTERNAL",			/* mech_name */
-  0,				/* max_ssf */
-  SASL_SEC_NOPLAINTEXT
-  | SASL_SEC_NODICTIONARY,	/* security_flags */
-  NULL,				/* required_prompts */
-  NULL,				/* glob_context */
-  &external_client_new,		/* mech_new */
-  &external_client_step,	/* mech_step */
-  NULL,				/* mech_dispose */
-  NULL,				/* mech_free */
-  NULL,				/* auth_create */
-  NULL				/* idle */
-};
-
-static int
-external_client_init(sasl_utils_t *utils,
-		     int max_version,
-		     int *out_version,
-		     const sasl_client_plug_t **pluglist,
-		     int *plugcount)
-{
-  if (!utils || !out_version || !pluglist || !plugcount)
-    return SASL_BADPARAM;
-  if (max_version != SASL_CLIENT_PLUG_VERSION)
-    return SASL_BADVERS;
-  *out_version = SASL_CLIENT_PLUG_VERSION;
-  *pluglist = &external_client_mech;
-  *plugcount = 1;
-  return SASL_OK;
-}
-
-typedef struct cmechanism
-{
-  int version;
-  const sasl_client_plug_t *plug;
-  void *library;
-
-  struct cmechanism *next;  
-} cmechanism_t;
-
-typedef struct sasl_client_conn {
-  sasl_conn_t base; /* parts common to server + client */
-
-  cmechanism_t *mech;
-  sasl_client_params_t *cparams;
-
-  char *serverFQDN;
-
-} sasl_client_conn_t;
-
-typedef struct cmech_list {
-  sasl_utils_t *utils; 
-
-  void *mutex;            /* mutex for this data */ 
-  cmechanism_t *mech_list; /* list of mechanisms */
-  int mech_length;       /* number of mechanisms */
-
-} cmech_list_t;
-
-
 static cmech_list_t *cmechlist; /* global var which holds the list */
 
 static sasl_global_callbacks_t global_callbacks;
 
+static int _sasl_client_active = 0;
+
 static int init_mechlist()
 {
+  cmechlist->mutex = sasl_MUTEX_ALLOC();
+  if(!cmechlist->mutex) return SASL_FAIL;
+  
   cmechlist->utils=_sasl_alloc_utils(NULL, &global_callbacks);
   if (cmechlist->utils==NULL)
     return SASL_NOMEM;
 
+  cmechlist->mech_list=NULL;
+  cmechlist->mech_length=0;
+
   return SASL_OK;
 }
 
@@ -261,54 +82,68 @@ static void client_done(void) {
   cmechanism_t *cprevm;
 
   cm=cmechlist->mech_list; /* m point to begging of the list */
-
   while (cm!=NULL)
   {
     cprevm=cm;
     cm=cm->next;
-    if (cprevm->library!=NULL)
-      _sasl_done_with_plugin(cprevm->library);
+
+    if (cprevm->plug->mech_free) {
+	cprevm->plug->mech_free(cprevm->plug->glob_context,
+				cmechlist->utils);
+    }
+
+    sasl_FREE(cprevm->plugname);
     sasl_FREE(cprevm);    
   }
-  sasl_MUTEX_DISPOSE(cmechlist->mutex);
+  sasl_MUTEX_FREE(cmechlist->mutex);
   _sasl_free_utils(&cmechlist->utils);
   sasl_FREE(cmechlist);
+
+  cmechlist = NULL;
+
+  _sasl_client_active = 0;
 }
 
-static int add_plugin(void *p, void *library) {
+int sasl_client_add_plugin(const char *plugname,
+			   sasl_client_plug_init_t *entry_point)
+{
   int plugcount;
-  const sasl_client_plug_t *pluglist;
+  sasl_client_plug_t *pluglist;
   cmechanism_t *mech;
-  sasl_client_plug_init_t *entry_point;
   int result;
   int version;
   int lupe;
 
-  entry_point = (sasl_client_plug_init_t *)p;
-
+  if(!plugname || !entry_point) return SASL_BADPARAM;
+  
   result = entry_point(cmechlist->utils, SASL_CLIENT_PLUG_VERSION, &version,
 		       &pluglist, &plugcount);
-  if (version != SASL_CLIENT_PLUG_VERSION)
-  {
-    VL(("Version conflict\n"));
-    result = SASL_FAIL;
-  }
+
   if (result != SASL_OK)
   {
-    VL(("entry_point failed\n"));
+    _sasl_log(NULL, SASL_LOG_WARN,
+	      "entry_point failed in sasl_client_add_plugin for %s",
+	      plugname);
     return result;
   }
 
+  if (version != SASL_CLIENT_PLUG_VERSION)
+  {
+    _sasl_log(NULL, SASL_LOG_WARN,
+	      "version conflict in sasl_client_add_plugin for %s", plugname);
+    return SASL_BADVERS;
+  }
+
   for (lupe=0;lupe< plugcount ;lupe++)
     {
       mech = sasl_ALLOC(sizeof(cmechanism_t));
       if (! mech) return SASL_NOMEM;
 
       mech->plug=pluglist++;
-      if (lupe==0)
-	mech->library = library;
-      else
-	mech->library = NULL;
+      if(_sasl_strdup(plugname, &mech->plugname, NULL) != SASL_OK) {
+	sasl_FREE(mech);
+	return SASL_NOMEM;
+      }
       mech->version = version;
       mech->next = cmechlist->mech_list;
       cmechlist->mech_list = mech;
@@ -350,6 +185,11 @@ client_idle(sasl_conn_t *conn)
 int sasl_client_init(const sasl_callback_t *callbacks)
 {
   int ret;
+  const add_plugin_list_t ep_list[] = {
+      { "sasl_client_plug_init", (void *)&sasl_client_add_plugin },
+      { "sasl_canonuser_init", (void *)&sasl_canonuser_add_plugin },
+      { NULL, NULL }
+  };
 
   _sasl_client_cleanup_hook = &client_done;
   _sasl_client_idle_hook = &client_idle;
@@ -360,27 +200,25 @@ int sasl_client_init(const sasl_callback_t *callbacks)
   cmechlist=sasl_ALLOC(sizeof(cmech_list_t));
   if (cmechlist==NULL) return SASL_NOMEM;
 
-  /* create mutex*/
-  cmechlist->mutex=sasl_MUTEX_NEW();
-
   /* load plugins */
   ret=init_mechlist();  
   if (ret!=SASL_OK)
     return ret;
 
-  cmechlist->mech_list=NULL;
-  cmechlist->mech_length=0;
-
-  add_plugin((void *) &external_client_init, NULL);
+  sasl_client_add_plugin("EXTERNAL", &external_client_init);
 
   ret = _sasl_common_init();
 
   if (ret == SASL_OK)
-      ret=_sasl_get_mech_list("sasl_client_plug_init",
-			      _sasl_find_getpath_callback(callbacks),
-			      _sasl_find_verifyfile_callback(callbacks),
-			      &add_plugin);
+      ret = _sasl_load_plugins(ep_list,
+			       _sasl_find_getpath_callback(callbacks),
+			       _sasl_find_verifyfile_callback(callbacks));
   
+  if (ret == SASL_OK) {
+      _sasl_client_active = 1;
+      ret = _sasl_build_mechlist();
+  }
+
   return ret;
 }
 
@@ -388,16 +226,20 @@ static void client_dispose(sasl_conn_t *pconn)
 {
   sasl_client_conn_t *c_conn=(sasl_client_conn_t *) pconn;
 
-  if (c_conn->mech && c_conn->mech->plug->mech_dispose)
-    c_conn->mech->plug->mech_dispose(c_conn->base.context,
+  if (c_conn->mech && c_conn->mech->plug->mech_dispose) {
+    c_conn->mech->plug->mech_dispose(pconn->context,
 				     c_conn->cparams->utils);
+  }
 
-  _sasl_free_utils(&c_conn->cparams->utils);
+  pconn->context = NULL;
 
-  if (c_conn->serverFQDN!=NULL)
-    sasl_FREE(c_conn->serverFQDN);
+  if (c_conn->serverFQDN)
+      sasl_FREE(c_conn->serverFQDN);
 
-  sasl_FREE(c_conn->cparams);
+  if (c_conn->cparams) {
+      _sasl_free_utils(&(c_conn->cparams->utils));
+      sasl_FREE(c_conn->cparams);
+  }
 
   _sasl_conn_dispose(pconn);
 }
@@ -405,6 +247,10 @@ static void client_dispose(sasl_conn_t *pconn)
 /* initialize a client exchange based on the specified mechanism
  *  service       -- registered name of the service using SASL (e.g. "imap")
  *  serverFQDN    -- the fully qualified domain name of the server
+ *  iplocalport   -- client IPv4/IPv6 domain literal string with port
+ *                    (if NULL, then mechanisms requiring IPaddr are disabled)
+ *  ipremoteport  -- server IPv4/IPv6 domain literal string with port
+ *                    (if NULL, then mechanisms requiring IPaddr are disabled)
  *  prompt_supp   -- list of client interactions supported
  *                   may also include sasl_getopt_t context & call
  *                   NULL prompt_supp = user/pass via SASL_INTERACT only
@@ -420,49 +266,63 @@ static void client_dispose(sasl_conn_t *pconn)
  *  SASL_NOMECH   -- no mechanism meets requested properties
  *  SASL_NOMEM    -- not enough memory
  */
-
 int sasl_client_new(const char *service,
 		    const char *serverFQDN,
+		    const char *iplocalport,
+		    const char *ipremoteport,
 		    const sasl_callback_t *prompt_supp,
-		    int secflags,
+		    unsigned flags,
 		    sasl_conn_t **pconn)
 {
   int result;
   sasl_client_conn_t *conn;
+  sasl_utils_t *utils;
+  
+  /* Remember, iplocalport and ipremoteport can be NULL and be valid! */
   if (!pconn || !service || !serverFQDN)
     return SASL_BADPARAM;
 
   *pconn=sasl_ALLOC(sizeof(sasl_client_conn_t));
-  if (*pconn==NULL) return SASL_NOMEM;
+  if (*pconn==NULL) {
+      _sasl_log(NULL, SASL_LOG_ERR,
+		"Out of memory allocating connection context");
+      return SASL_NOMEM;
+  }
+  memset(*pconn, 0, sizeof(sasl_client_conn_t));
 
   (*pconn)->destroy_conn = &client_dispose;
-  result = _sasl_conn_init(*pconn, service, secflags,
-			   &client_idle, NULL,
-			   prompt_supp, &global_callbacks);
-  if (result != SASL_OK) return result;
-  
-  conn = (sasl_client_conn_t *)*pconn;
 
+  conn = (sasl_client_conn_t *)*pconn;
+  
   conn->mech = NULL;
 
   conn->cparams=sasl_ALLOC(sizeof(sasl_client_params_t));
-  if (conn->cparams==NULL) return SASL_NOMEM;
-
-  conn->cparams->utils=_sasl_alloc_utils(*pconn, &global_callbacks);
-  if (conn->cparams->utils==NULL)
-    return SASL_NOMEM;
-
-  conn->cparams->utils->conn= *pconn;
+  if (conn->cparams==NULL) 
+      MEMERROR(*pconn);
+  memset(conn->cparams,0,sizeof(sasl_client_params_t));
 
+  result = _sasl_conn_init(*pconn, service, flags, SASL_CONN_CLIENT,
+			   &client_idle, serverFQDN,
+			   iplocalport, ipremoteport,
+			   prompt_supp, &global_callbacks);
+  if (result != SASL_OK) RETURN(*pconn, result);
+  
+  utils=_sasl_alloc_utils(*pconn, &global_callbacks);
+  if (utils==NULL)
+      MEMERROR(*pconn);
+  
+  utils->conn= *pconn;
+  conn->cparams->utils = utils;
+  conn->cparams->canon_user = &_sasl_canon_user;
+  
   result = _sasl_strdup(serverFQDN, &conn->serverFQDN, NULL);
-
-  if (result == SASL_OK) return SASL_OK;
+  if(result == SASL_OK) return SASL_OK;
 
   /* result isn't SASL_OK */
-
   _sasl_conn_dispose(*pconn);
   sasl_FREE(*pconn);
   *pconn = NULL;
+  _sasl_log(NULL, SASL_LOG_ERR, "Out of memory in sasl_client_new");
   return result;
 }
 
@@ -516,12 +376,10 @@ static int have_prompts(sasl_conn_t *conn,
  * Apps should be encouraged to simply use space or comma space
  * though
  */
-
 int sasl_client_start(sasl_conn_t *conn,
-		      const char *list,
-		      sasl_secret_t *secret,
+		      const char *mechlist,
 		      sasl_interact_t **prompt_need,
-		      char **clientout,
+		      const char **clientout,
 		      unsigned *clientoutlen,
 		      const char **mech)
 {
@@ -533,57 +391,44 @@ int sasl_client_start(sasl_conn_t *conn,
     sasl_ssf_t bestssf = 0, minssf = 0;
     int result;
 
-    /* verify parameters */
-    if (list == NULL)
-    {
-	return SASL_BADPARAM;
-    }
+    if (!conn) return SASL_BADPARAM;
 
-    VL(("in sasl_client_start\n"));
+    /* verify parameters */
+    if (mechlist == NULL)
+	PARAMERROR(conn);
 
     /* if prompt_need != NULL we've already been here
        and just need to do the continue step again */
 
     /* do a step */
+    /* FIXME: Hopefully they only give us our own prompt_need back */
     if (prompt_need && *prompt_need != NULL) {
-	result = c_conn->mech->plug->mech_step(conn->context,
-					       c_conn->cparams,
-					       NULL,
-					       0,
-					       prompt_need,
-					       clientout, (int *) clientoutlen,
-					       &conn->oparams);
-
-	return result;
+	goto dostep;
     }
 
-    /* set secret */
-    conn->secret=secret;
-
-    /* Get app's desired sec props */
-    if (conn->props.min_ssf < conn->external.ssf) {
+    if(conn->props.min_ssf < conn->external.ssf) {
 	minssf = 0;
     } else {
 	minssf = conn->props.min_ssf - conn->external.ssf;
     }
 
     /* parse mechlist */
-    VL(("mech list from server is %s\n", list));
-    list_len = strlen(list);
+    list_len = strlen(mechlist);
+
     while (pos<list_len)
     {
 	place=0;
-	while ((pos<list_len) && (isalnum((unsigned char)list[pos])
-				  || list[pos] == '_'
-				  || list[pos] == '-')) {
-	    name[place]=list[pos];
+	while ((pos<list_len) && (isalnum((unsigned char)mechlist[pos])
+				  || mechlist[pos] == '_'
+				  || mechlist[pos] == '-')) {
+	    name[place]=mechlist[pos];
 	    pos++;
 	    place++;
 	    if (SASL_MECHNAMEMAX < place) {
 		place--;
-		while(pos<list_len && (isalnum((unsigned char)list[pos])
-				       || list[pos] == '_'
-				       || list[pos] == '-'))
+		while(pos<list_len && (isalnum((unsigned char)mechlist[pos])
+				       || mechlist[pos] == '_'
+				       || mechlist[pos] == '-'))
 		    pos++;
 	    }
 	}
@@ -592,8 +437,6 @@ int sasl_client_start(sasl_conn_t *conn,
 
 	if (! place) continue;
 
-	VL(("Considering mech %s\n",name));
-
 	/* foreach in server list */
 	for (m = cmechlist->mech_list; m != NULL; m = m->next) {
 	    /* is this the mechanism the server is suggesting? */
@@ -614,6 +457,12 @@ int sasl_client_start(sasl_conn_t *conn,
 		break;
 	    }
 
+	    /* Can we meet it's features? */
+	    if ((m->plug->features & SASL_FEAT_NEEDSERVERFQDN)
+		&& !conn->serverFQDN) {
+		break;
+	    }
+	    
 #ifdef PREFER_MECH
 	    if (strcasecmp(m->plug->mech_name, PREFER_MECH) &&
 		bestm && m->plug->max_ssf <= bestssf) {
@@ -628,7 +477,6 @@ int sasl_client_start(sasl_conn_t *conn,
 	    }
 #endif
 
-	    VL(("Best mech so far: %s\n", m->plug->mech_name));
 	    if (mech) {
 		*mech = m->plug->mech_name;
 	    }
@@ -639,7 +487,7 @@ int sasl_client_start(sasl_conn_t *conn,
     }
 
     if (bestm == NULL) {
-	VL(("No worthy mechs found\n"));
+	sasl_seterror(conn, 0, "No worthy mechs found");
 	result = SASL_NOMECH;
 	goto done;
     }
@@ -652,20 +500,32 @@ int sasl_client_start(sasl_conn_t *conn,
     c_conn->mech = bestm;
 
     /* init that plugin */
-    c_conn->mech->plug->mech_new(NULL,
-				 c_conn->cparams,
-				 &(conn->context));
+    result = c_conn->mech->plug->mech_new(NULL,
+					  c_conn->cparams,
+					  &(conn->context));
+    if(result != SASL_OK) goto done;
+
+    /* do a step -- but only if we can do a client-send-first */
+ dostep:
+    if(c_conn->mech->plug->features & SASL_FEAT_INTERNAL_CLIENT_FIRST) {
+	/* The plugin handles client-first internally */
+	result = sasl_client_step(conn, NULL, 0, prompt_need,
+				  clientout, clientoutlen);
+    } else if(clientout) {
+	if(c_conn->mech->plug->features & SASL_FEAT_WANT_CLIENT_FIRST) {
+	    result = sasl_client_step(conn, NULL, 0, prompt_need,
+				      clientout, clientoutlen);
+	} else {
+	    *clientout = NULL;
+	    *clientoutlen = 0;
+	    result = SASL_CONTINUE;
+	}
+    }
+    else
+	result = SASL_CONTINUE;
 
-    /* do a step */
-    result = c_conn->mech->plug->mech_step(conn->context,
-					 c_conn->cparams,
-					 NULL,
-					 0,
-					 prompt_need,
-					 clientout, (int *) clientoutlen,
-					 &conn->oparams);    
  done:
-    return result;
+    RETURN(conn, result);
 }
 
 /* do a single authentication step.
@@ -686,15 +546,17 @@ int sasl_client_step(sasl_conn_t *conn,
 		     const char *serverin,
 		     unsigned serverinlen,
 		     sasl_interact_t **prompt_need,
-		     char **clientout,
+		     const char **clientout,
 		     unsigned *clientoutlen)
 {
   sasl_client_conn_t *c_conn= (sasl_client_conn_t *) conn;
   int result;
 
+  if(!conn) return SASL_BADPARAM;
+
   /* check parameters */
   if ((serverin==NULL) && (serverinlen>0))
-    return SASL_BADPARAM;
+      PARAMERROR(conn);
 
   /* do a step */
   result = c_conn->mech->plug->mech_step(conn->context,
@@ -704,53 +566,188 @@ int sasl_client_step(sasl_conn_t *conn,
 					 prompt_need,
 					 clientout, (int *)clientoutlen,
 					 &conn->oparams);
+
+  if (result == SASL_OK) {
+      /* So we're done on this end, but if both
+       * 1. the mech does server-send-last
+       * 2. the protocol does not
+       * we need to return no data */
+      if(!(conn->flags & SASL_SUCCESS_DATA)
+	 && (c_conn->mech->plug->features & SASL_FEAT_WANT_SERVER_LAST)) {
+	  *clientout = "";
+	  *clientoutlen = 0;
+      }
+      
+      if(!conn->oparams.maxoutbuf) {
+	  conn->oparams.maxoutbuf = conn->props.maxbufsize;
+      }
+
+      if(conn->oparams.user == NULL || conn->oparams.authid == NULL) {
+	  sasl_seterror(conn, 0,
+			"mech did not call canon_user for both authzid and authid");
+	  result = SASL_BADPROT;
+      }
+  }
+  
+
+  RETURN(conn,result);
+}
+
+/* returns the length of all the mechanisms
+ * added up 
+ */
+
+static unsigned mech_names_len()
+{
+  cmechanism_t *listptr;
+  unsigned result = 0;
+
+  for (listptr = cmechlist->mech_list;
+       listptr;
+       listptr = listptr->next)
+    result += strlen(listptr->plug->mech_name);
+
   return result;
 }
 
-/* Set connection secret based on passphrase
- *  may be used in SASL_CB_PASS callback
- * input:
- *  user          -- username
- *  pass          -- plaintext passphrase with NUL sentinel
- *  passlen       -- 0 = strlen(pass)
- * out:
- *  prompts       -- if non-NULL, SASL_CB_PASS item filled in
- *  keepcopy      -- set to copy of secret if non-NULL
- * returns:
- *  SASL_OK       -- success
- *  SASL_NOMEM    -- failure
- */
 
-int sasl_client_auth(sasl_conn_t *conn __attribute__((unused)),
-		     const char *user __attribute__((unused)),
-		     const char *pass __attribute__((unused)), 
-		     unsigned passlen __attribute__((unused)),
-		     sasl_interact_t *prompts __attribute__((unused)),
-		     sasl_secret_t **keepcopy __attribute__((unused)))
+int _sasl_client_listmech(sasl_conn_t *conn,
+			  const char *prefix,
+			  const char *sep,
+			  const char *suffix,
+			  const char **result,
+			  unsigned *plen,
+			  int *pcount)
 {
-  /* XXX needs to be filled in */
-  return SASL_FAIL;
+    cmechanism_t *m=NULL;
+    sasl_ssf_t minssf = 0;
+    int ret;
+    int resultlen;
+    int flag;
+    const char *mysep;
+
+    if(_sasl_client_active == 0) return SASL_NOTINIT;
+    if (!conn) return SASL_BADPARAM;
+    if(conn->type != SASL_CONN_CLIENT) PARAMERROR(conn);
+    
+    if (! result)
+	PARAMERROR(conn);
+    
+    if (plen != NULL)
+	*plen = 0;
+    if (pcount != NULL)
+	*pcount = 0;
+
+    if (sep) {
+	mysep = sep;
+    } else {
+	mysep = " ";
+    }
+
+    if(conn->props.min_ssf < conn->external.ssf) {
+	minssf = 0;
+    } else {
+	minssf = conn->props.min_ssf - conn->external.ssf;
+    }
+
+    if (! cmechlist || cmechlist->mech_length <= 0)
+	INTERROR(conn, SASL_NOMECH);
+
+    resultlen = (prefix ? strlen(prefix) : 0)
+	+ (strlen(mysep) * (cmechlist->mech_length - 1))
+	+ mech_names_len()
+	+ (suffix ? strlen(suffix) : 0)
+	+ 1;
+    ret = _buf_alloc(&conn->mechlist_buf,
+		     &conn->mechlist_buf_len, resultlen);
+    if(ret != SASL_OK) MEMERROR(conn);
+
+    if (prefix)
+	strcpy (conn->mechlist_buf,prefix);
+    else
+	*(conn->mechlist_buf) = '\0';
+
+    flag = 0;
+    for (m = cmechlist->mech_list; m != NULL; m = m->next) {
+	    /* do we have the prompts for it? */
+	    if (!have_prompts(conn, m->plug))
+		continue;
+
+	    /* is it strong enough? */
+	    if (minssf > m->plug->max_ssf)
+		continue;
+
+	    /* does it meet our security properties? */
+	    if (((conn->props.security_flags ^ m->plug->security_flags)
+		 & conn->props.security_flags) != 0) {
+		continue;
+	    }
+
+	    /* Can we meet it's features? */
+	    if ((m->plug->features & SASL_FEAT_NEEDSERVERFQDN)
+		&& !conn->serverFQDN) {
+		continue;
+	    }
+
+	    /* Okay, we like it, add it to the list! */
+
+	    if (pcount != NULL)
+		(*pcount)++;
+
+	    /* print seperator */
+	    if (flag) {
+		strcat(conn->mechlist_buf, mysep);
+	    } else {
+		flag = 1;
+	    }
+	    
+	    /* now print the mechanism name */
+	    strcat(conn->mechlist_buf, m->plug->mech_name);
+    }
+    
+  if (suffix)
+      strcat(conn->mechlist_buf,suffix);
+
+  if (plen!=NULL)
+      *plen=strlen(conn->mechlist_buf);
+
+  *result = conn->mechlist_buf;
+
+  return SASL_OK;
 }
 
-/* erase & dispose of a sasl_secret_t
- *  calls free utility last set by sasl_set_alloc
- */
-
-void sasl_free_secret(sasl_secret_t **secret)
+sasl_string_list_t *_sasl_client_mechs(void) 
 {
-  size_t lup;
+  cmechanism_t *listptr;
+  sasl_string_list_t *retval = NULL, *next=NULL;
 
-  VL(("trying to free secret\n"));
+  if(!_sasl_client_active) return NULL;
 
-  if (secret==NULL) return;
-  if (*secret==NULL) return;
+  /* make list */
+  for (listptr = cmechlist->mech_list; listptr; listptr = listptr->next) {
+      next = sasl_ALLOC(sizeof(sasl_string_list_t));
 
-  /* overwrite the memory */
-  for (lup=0;lup<(*secret)->len;lup++)
-    (*secret)->data[lup]='\0';
+      if(!next && !retval) return NULL;
+      else if(!next) {
+	  next = retval->next;
+	  do {
+	      sasl_FREE(retval);
+	      retval = next;
+	      next = retval->next;
+	  } while(next);
+	  return NULL;
+      }
+      
+      next->d = listptr->plug->mech_name;
 
-  sasl_FREE(*secret);
+      if(!retval) {
+	  next->next = NULL;
+	  retval = next;
+      } else {
+	  next->next = retval;
+	  retval = next;
+      }
+  }
 
-  *secret=NULL;
+  return retval;
 }
-
