@@ -28,11 +28,10 @@
  * END COPYRIGHT */
 
 #ifdef __GNUC__
-#ident "$Id: auth_krb4.c,v 1.3 2001/07/24 17:11:28 rjs3 Exp $"
+#ident "$Id: auth_krb4.c,v 1.5 2001/12/18 02:34:32 rjs3 Exp $"
 #endif
 
 /* PUBLIC DEPENDENCIES */
-#include <config.h>
 #include <unistd.h>
 #include "mechanisms.h"
 
@@ -122,7 +121,7 @@ auth_krb4_init (
     }
 
     /* Make sure it's not a symlink. */
-    if (sb.st_mode & S_IFLNK) {
+    if (S_ISLNK(sb.st_mode)) {
         syslog(LOG_ERR, "auth_krb4_init: %s: is a symbolic link", tf_dir);
 	free(tf_dir);
 	tf_dir = NULL;
@@ -151,7 +150,9 @@ char *					/* R: allocated response string */
 auth_krb4 (
   /* PARAMETERS */
   const char *login,			/* I: plaintext authenticator */
-  const char *password			/* I: plaintext password */
+  const char *password,			/* I: plaintext password */
+  const char *service __attribute__((unused)),
+  const char *realm_in __attribute__((unused))
   /* END PARAMETERS */
   )
 {
@@ -159,6 +160,7 @@ auth_krb4 (
     char aname[ANAME_SZ];		/* Kerberos principal */
     char inst[INST_SZ];			/* Kerberos instance (default: imap) */
     char realm[REALM_SZ];		/* Kerberos realm to authenticate in */
+    char pidstr[128];
     int rc;				/* return code */
     static char *tf_name;		/* Ticket file name */
     /* END VARIABLES */
@@ -179,7 +181,8 @@ auth_krb4 (
      * NOTE: these ticket files are not used by us. The are created
      * as a side-effect of calling krb_get_pw_in_tkt.
      */
-    tf_name = malloc(strlen(tf_dir) + strlen(login) + 2);
+    snprintf(pidstr, sizeof(pidstr), "%d", getpid());
+    tf_name = malloc(strlen(tf_dir) + strlen(login) + strlen(pidstr) + 2);
     if (tf_name == NULL) {
 	syslog(LOG_ERR, "auth_krb4: malloc(tf_name) failed");
 	return strdup("NO saslauthd internal error");
@@ -187,6 +190,7 @@ auth_krb4 (
     }
     strcpy(tf_name, tf_dir);
     strcat(tf_name, login);
+    strcat(tf_name, pidstr);
     krb_set_tkt_string(tf_name);
     
     strncpy(aname, login, ANAME_SZ-1);
@@ -231,7 +235,9 @@ auth_krb4 (
 char *
 auth_krb4 (
   const char *login __attribute__((unused)),
-  const char *password __attribute__((unused))
+  const char *password __attribute__((unused)),
+  const char *service __attribute__((unused)),
+  const char *realm __attribute__((unused))
   )
 {
     return NULL;
